/*   -*- c++ -*-
$Header: /home/auspex/h/bair/phelps/www/java/zip/RCS/Pinhead.java,v 1.7 1996/02/22 02:58:43 phelps Exp phelps $

	Pinhead - a .zip-enabled ClassLoader

	Tom Phelps

	15 Feb	initiated (~100 lines, 2K characters)
	16      it lives! can load classes from zip files!
	17      load and search multiple Zips, free zip when load all its classes
	18      embedding ZipApplet, package support: '.'=>'/'
	20      package as a zip its ownself, for placement in Netscape CLASSPATH

*/

//package EDU.Berkeley.CS.Phelps.Zip;

import java.lang.*;
import java.util.*;
import java.awt.Image;
import java.io.*;
import java.applet.*;
import java.net.URL;
import java.net.MalformedURLException;
//import EDU.Berkeley.CS.Phelps.Zip.ZipApplet;
import ZipApplet;
//import EDU.Berkeley.CS.Phelps.Util.tapUtils;
import tapUtils;


// put all Java specifics in Pinhead

class PinheadClassLoader extends ClassLoader {
  final boolean debug = false;
  public static boolean showstats = false;		// turn on to show classes loaded from zip file

  final String classSuffix = ".class";

  Hashtable cache = new Hashtable();
  Zippy zippy[];
  int classesLoaded[];		// track number of classes loaded from zip, free zip when got them all
  String zippyname[];

  /*
   * CONSTRUCTORS
   */

  public PinheadClassLoader(URL base, String zip) {
	// multiple .zip files separated by colons
	StringTokenizer st = new StringTokenizer(zip, ":");

	zippy = new Zippy[st.countTokens()];
	classesLoaded = new int[zippy.length];
	zippyname = new String[zippy.length];

	String zipname;
	InputStream in;
	URL url;
	for (int i=0; i<zippy.length; i++) {
	  try {
		zipname = st.nextToken();
		if (!zipname.endsWith(Zippy.suffix)) zipname+=Zippy.suffix;
		url = new URL(base,zipname);
		in = url.openStream();
		zippy[i] = new Zippy(in);
		zippyname[i] = zipname;
		classesLoaded[i] = 0;
		in.close();
	  } catch (MalformedURLException e) {
		System.out.println("malformed URL: "+e);
	  } catch (IOException e) {
		System.out.println("can't open stream: "+e);
	  }
	}
  }

  boolean valid(Zippy zip) {
	// Java-specific diagnostics
	// (no compression=>in Zippy, cafe babe=>in ClassLoader)
	int filecnt = zip.dirhook.num_entries_centrl_dir_ths_disk;
	ZipDirEntry z;
	for (int i=0; i<filecnt; i++) {
		 z = zip.direntry[i];
	}

	return true;
  }



  /*
   * FUNCTIONALITY
   */

  public Class loadClass(String name, boolean resolve) throws ClassNotFoundException {

	if (debug) System.out.println("loadClass on "+name+" resolve="+resolve);

	String findme = name.replace('.','/');
	if (!findme.endsWith(classSuffix)) findme += classSuffix;
	Class newclass;

	if ((newclass=(Class)cache.get(name))==null) {
//	  if (showstats) System.out.println("already defined: "+newclass.getName());
//	} else {
	  if (showstats) System.out.println("need to load "+name);

	  // search all zip files, classes listed earlier get higher priority
	  ZipDirEntry z=null;
	  Zippy zip=null;
	  for (int i=0; i<zippy.length; i++) {
		zip = zippy[i]; if (zip==null) continue;
		z = zip.findName(findme);
		if (z!=null) {
		  tapUtils.assert(isJavaClass(zip.data,z.getDataOffset()), "not a Java class: "+z.filename);
		  if (showstats) System.out.println("fetched "+name+" from zip");
		  newclass = defineClass(zip.data,z.getDataOffset(),z.uncompressed_size);
		  classesLoaded[i]++;
		  if (classesLoaded[i]==zip.dirhook.total_entries_central_dir) {
			if (showstats) System.out.println("*** FREEING "+zippyname[i]+" ***");
			zippy[i]=null;
		  }
		  break;
		}
	  }
	  if (newclass==null) {
		//System.out.println("loading "+name+" from system classes");
		newclass = findSystemClass(name);	// throws ClassNotFoundException
		//System.out.println("loaded "+name+" from system classes");
	  }

	  cache.put(name, newclass);
	}

	if (resolve) resolveClass(newclass);
	if (resolve && debug) { System.out.println("resolved class "+name); }
//System.out.println("returning "+newclass.getName());
	return newclass;
  }



  /*
   * DIAGNOSTICS
   */

  public static boolean isJavaClass(byte data[], int offset) {
	byte javasig[] = { (byte)0xca,(byte)0xfe,(byte)0xba,(byte)0xbe };	// cafe babe
	return tapUtils.bytesidentical(javasig,0, data,offset, javasig.length);
  }
}


// load real Applet as a shadow of Pinhead
public class Pinhead extends Applet {
  //Applet realapp=null;
  ZipApplet realapp=null;
  public boolean debug = false;
  Image img=null;
  AudioClip clip;

  public void init() {
/* by the time we get here, a SecurityManager is already set
	try {
	  System.setSecurityManager(new CLSecurityManager());
	} catch (Exception e) { System.out.println(e); }
*/
/* this works fine
	try {
	  img = getImage(new URL("http://www.cs.berkeley.edu/~phelps/java/pix/Duke.gif"));
	  System.out.println("*** got image");
	} catch (Exception e) {}
	try {
	  URL url = new URL(getCodeBase(), "train.au");
	  clip = getAudioClip(url);
	  System.out.println("*** got audio");
	} catch (Exception e) {
	  System.out.println("couldn't load audio");
	}
*/

	String zip = getParameter("zip");
	String applet = getParameter("applet");
	String codebase = getParameter("codebase");
//	if (codebase==null) codebase = getCodeBase().toString();
	if ("true".equals(getParameter("showstats"))) PinheadClassLoader.showstats=true;
	if ("true".equals(getParameter("debug"))) debug=true;

	PinheadClassLoader loader = new PinheadClassLoader(getDocumentBase(), zip);	// not clear what getCodeBase points to -- null?
	try {
	  //realapp = (Applet)loader.loadClass(applet,true).newInstance();
	  //realapp.setStub(getStub(stub));	// link to outside world -- no getStub(), even protected!
	  realapp = (ZipApplet)loader.loadClass(applet,true).newInstance();
	  realapp.setApplet(this);
	  add("Center", realapp);	// embed real thing
	} catch (ClassNotFoundException e) {
	  System.out.println(e);
	  throw new Error();
	} catch (Exception e) {
	  System.out.println("can't instantiate: "+e);
	  throw new Error();
	}
	//this = realapp;  -- invalid LHS

	if (debug) System.out.println(realapp.getClass().getName()+" init");
//	realapp.resize(size().width, size().height);
	realapp.init();
  }

  // redirect Applet callbacks
  public void start() {
	if (debug) System.out.println(realapp.getClass().getName()+" start");
	realapp.start();
  }
  public void stop() {
	if (debug) System.out.println(realapp.getClass().getName()+" stop");
	realapp.stop();
  }
  public void destroy() {
	if (debug) System.out.println(realapp.getClass().getName()+" destroy");
	realapp.destroy();
  }
  public String getAppletInfo() { return realapp.getAppletInfo(); }
  public String[][] getParameterInfo() { return realapp.getParameterInfo(); }

  public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
System.out.println("made it to imageUpdate");
	return realapp.imageUpdate(img,infoflags,x,y,width,height);
  }

  // debugging
  public Image getImage(URL url) {
	System.out.println("in Pinhead");
	Object i=null;
	try {
//	  i = getAppletContext().getImage(url);
	  i = super.getImage(url);
	} catch (Exception e) { System.out.println("*** BOOM: "+e); }
//	System.out.println("return class = "+i.getClass().getName());
	System.out.println("returning from Pinhead");
	return (Image)i;
  }
}
