/*   -*- c++ -*-
$Header: /home/auspex/h/bair/phelps/www/java/zip/RCS/Zippy.java,v 1.6 1996/02/22 02:56:30 phelps Exp phelps $

	Zippy - generic zip wrangler, written on a need-to-implement basis for Pinhead
            (for now, single-disk, uncompressed files, Zip 1.0 format)

	Tom Phelps, after zipinfo by Greg Roelofs (also known as Cave Newt)

	15 Feb	initiated (~400 lines, 13K characters).  interrogates simple .zip formats
	16      added findName(String name)

TO DO
add more diagnostics -- gotta be single-disk, uncompressed

*/

//package EDU.Berkeley.CS.Phelps.Zip;

import java.lang.*;
import java.io.*;
//import EDU.Berkeley.CS.Phelps.Util.tapUtils;
import tapUtils;


// big endian i/o

class ZipDataInputStream extends FilterInputStream {
  ZipDataInputStream(InputStream in) {
	   super(in);
  }

  public final byte readByte() throws IOException {
	int ch = in.read();
	return (byte)ch;
  }

  public final short readShort() throws IOException {
	   InputStream in = this.in;
	   int ch1 = in.read();
	   int ch2 = in.read();
	   if ((ch1 | ch2) < 0)
			throw new EOFException();
	   return (short)((ch2 << 8) + (ch1 << 0));
  }

  public final int readInt() throws IOException {
	   InputStream in = this.in;
	   int ch1 = in.read();
	   int ch2 = in.read();
	   int ch3 = in.read();
	   int ch4 = in.read();
	   if ((ch1 | ch2 | ch3 | ch4) < 0)
			throw new EOFException();
	   return ((ch4 << 24) + (ch3 << 16) + (ch2 << 8) + (ch1 << 0));
  }

  public final String readString(int len) throws IOException {
	   InputStream in = this.in;
	   if (len==0) return "";
	   StringBuffer sb = new StringBuffer(len);
	   int ch;
	   while (len-->0) {
			ch = in.read();
			if (ch<0) throw new EOFException();
			sb.append((char)ch);
	   }
	   return sb.toString();
  }
}



class ZipDirEntry {
  // file format
  public final static int recsize = 42+4;
  public final static byte signature[] = { 'P','K',0x01,0x02 };

  byte version_made_by[] = new byte[2];
  byte version_needed_to_extract[] = new byte[2];
  short general_purpose_bit_flag;
  short compression_method;
  short last_mod_file_time;
  short last_mod_file_date;
  int crc32;
  int compressed_size;
  public int uncompressed_size;		// change then when/if add support for compression
  short filename_length;
  short extra_field_length;
  short file_comment_length;
  short disk_number_start;
  short internal_file_attributes;
  int external_file_attributes;
  int relative_offset_local_header;

  String filename;
  String extra_field;
  String file_comment;
  private int dataoffset=-1;


  // instance data
  public byte data[];
  private int offset;
  private int size;
  

  ZipDirEntry(byte data[], int offset) {	// in lieu of pointers...
	this.data = data; this.offset = offset;
	try {
		ZipDataInputStream in = new ZipDataInputStream(new ByteArrayInputStream(data,offset,recsize));

		byte sig[] = new byte[4];
		in.read(sig,0,signature.length);
		tapUtils.assert(tapUtils.bytesidentical(signature,0, sig,0, signature.length), "corrupted directory entry signature");

		version_made_by[0] = in.readByte(); version_made_by[1] = in.readByte();
		version_needed_to_extract[0] = in.readByte(); version_needed_to_extract[1] = in.readByte();
		general_purpose_bit_flag = in.readShort();
		compression_method = in.readShort();
		last_mod_file_time = in.readShort();
		last_mod_file_date = in.readShort();
		crc32 = in.readInt();
		compressed_size = in.readInt();
		uncompressed_size = in.readInt();
		filename_length = in.readShort();
		extra_field_length = in.readShort();
		file_comment_length = in.readShort();
		disk_number_start = in.readShort();
		internal_file_attributes = in.readShort();
		external_file_attributes = in.readInt();
		relative_offset_local_header = in.readInt();
		in.close();

		// now read filename and comment
		int strlen = filename_length+extra_field_length+file_comment_length;
		size = recsize+strlen;
		in = new ZipDataInputStream(new ByteArrayInputStream(data,offset+recsize, strlen));
		filename = in.readString(filename_length);
		extra_field = in.readString(extra_field_length);
		file_comment = in.readString(file_comment_length);
		in.close();

    } catch (Exception e) {
		 System.out.println("trouble reading central directory entry: "+e);
		 return;
	}
  }


  /*
   * ACCESSORS
   */

  public int size() {
	return size;
  }

  public int getDataOffset() {
	// make checks with local header and set dataoffset
	if (dataoffset==-1) {
		 ZipLocalHeader l = new ZipLocalHeader(data,relative_offset_local_header,this);
		 dataoffset = l.getDataOffset();
	}
	return dataoffset;
  }

  public String toString() {
	// augment with first several bytes of data
	return method[compression_method]+" "+filename+" "+uncompressed_size+", "+relative_offset_local_header+"=> "+tapUtils.bytestoString(data,getDataOffset(),4+5)+" ... "+tapUtils.bytestoString(data,getDataOffset()+uncompressed_size-5,5);
  }


  /*
   * DIAGNOSTICS
   */

  static final String method[] = {
	   "Stored", "Shrunk", "Reduce1", "Reduce2", "Reduce3", "Reduce4",
	   "Implode", "Token", "Deflate", "UNKNOWN"
  };
  public void dump() {
	System.out.println("DIR ENTRY @ "+Integer.toString(offset,8));
	System.out.println("version_made_by = "+version_made_by[0]+"."+version_made_by[1]);
	System.out.println("version_needed_to_extract = "+version_needed_to_extract[0]+"."+version_needed_to_extract[1]);
	System.out.println("general_purpose_bit_flag = "+general_purpose_bit_flag);
	System.out.println("compression_method = "+method[compression_method]);
	System.out.println("last_mod_file_time = "+last_mod_file_time);
	System.out.println("last_mod_file_date = "+last_mod_file_date);
	System.out.println("crc32 = "+crc32);
	System.out.println("compressed_size = "+compressed_size);
	System.out.println("uncompressed_size = "+uncompressed_size);
	System.out.println("filename_length = "+filename_length+" ("+filename+")");
	System.out.println("extra_field_length = "+extra_field_length+" ("+extra_field+")");
	System.out.println("file_comment_length = "+file_comment_length+" ("+file_comment+")");
	System.out.println("disk_number_start = "+disk_number_start);
	System.out.println("internal_file_attributes = "+Integer.toString(internal_file_attributes,16));
	System.out.println("external_file_attributes = "+Integer.toString(external_file_attributes,16));
	System.out.println("relative_offset_local_header = "+relative_offset_local_header);
  }
}



class ZipDirHook {

  // file format
  public final static int recsize = 18+4;
  public final static byte signature[] = { 'P','K',0x05,0x06 };

  short number_this_disk;
  short num_disk_with_start_central_dir;
  short num_entries_centrl_dir_ths_disk;
  public short total_entries_central_dir;
  int size_central_directory;
  int offset_start_central_directory;
  short zipfile_comment_length;


  ZipDirHook(byte data[], int offset) {	// in lieu of pointers...
	try {
		ZipDataInputStream in = new ZipDataInputStream(new ByteArrayInputStream(data,offset,recsize));

		byte sig[] = new byte[4];
		in.read(sig,0,signature.length);
		tapUtils.assert(tapUtils.bytesidentical(signature,0, sig,0, signature.length), "corrupted directory entry signature");

		number_this_disk = in.readShort();
		num_disk_with_start_central_dir = in.readShort();
		num_entries_centrl_dir_ths_disk = in.readShort();
		total_entries_central_dir = in.readShort();
		size_central_directory = in.readInt();
		offset_start_central_directory = in.readInt();
		zipfile_comment_length = in.readShort();
		in.close();
    } catch (Exception e) {
		 System.out.println("trouble reading central directory end: "+e);
		 return;
	}
  }


  /*
   * DIAGNOSTICS
   */

  public void dump() {
	System.out.println("DIR HOOK");
	System.out.println("number_this_disk = "+number_this_disk);
	System.out.println("number_this_disk_with_start_central_dir = "+number_this_disk);
	System.out.println("num_entries_centrl_dir_ths_disk = "+num_entries_centrl_dir_ths_disk);
	System.out.println("total_entries_central_dir = "+total_entries_central_dir);
	System.out.println("size_central_directory = "+size_central_directory);
	System.out.println("offset_start_central_directory = "+offset_start_central_directory);
	System.out.println("zipfile_comment_length = "+zipfile_comment_length);
  }
}


// used temporarily to verify header and retrieve pointer to class data
class ZipLocalHeader {

  // file format
  public final static int recsize = 26+4;
  public final static byte signature[] = { 'P','K',0x03,0x04 };

  byte version_needed_to_extract[] = new byte[2];
  short general_purpose_bit_flag;
  short compression_method;
  short last_mod_file_time;
  short last_mod_file_date;
  long crc32;
  long compressed_size;
  long uncompressed_size;
  short filename_length;
  short extra_field_length;


  // instance variables
  byte data[];
  int offset;
  private int size;


  ZipLocalHeader(byte data[], int offset, ZipDirEntry central) {
	this.data = data; this.offset = offset;

	String filename;
	String extra_field;
	try {
		ZipDataInputStream in = new ZipDataInputStream(new ByteArrayInputStream(data,offset,recsize+central.filename_length));

		byte sig[] = new byte[4];
		in.read(sig,0,signature.length);
		tapUtils.assert(tapUtils.bytesidentical(signature,0, sig,0, signature.length), "corrupted directory entry signature");

		version_needed_to_extract[0] = in.readByte(); version_needed_to_extract[1] = in.readByte();
		general_purpose_bit_flag = in.readShort();
		compression_method = in.readShort();
		last_mod_file_time = in.readShort();
		last_mod_file_date = in.readShort();
		crc32 = in.readInt();
		compressed_size = in.readInt();
		uncompressed_size = in.readInt();
		filename_length = in.readShort();
		extra_field_length = in.readShort();

		filename = in.readString(filename_length);
		extra_field = in.readString(extra_field_length);
		in.close();
    } catch (Exception e) {
		 System.out.println("trouble reading local directory entry: "+e);
		 return;
	}

	size = recsize + filename_length + extra_field_length;

	// verify that all fields match up
	String beep = "local field doesn't match central directory field";
	tapUtils.assert(general_purpose_bit_flag==central.general_purpose_bit_flag, beep);
	tapUtils.assert(compression_method==central.compression_method, beep);
	tapUtils.assert(last_mod_file_time==central.last_mod_file_time, beep);
	tapUtils.assert(last_mod_file_date==central.last_mod_file_date, beep);
	tapUtils.assert(crc32==central.crc32, beep);
	tapUtils.assert(compressed_size==central.compressed_size, beep);
	tapUtils.assert(uncompressed_size==central.uncompressed_size, beep);
	tapUtils.assert(filename_length==central.filename_length, beep);
	tapUtils.assert(extra_field_length==central.extra_field_length, beep);
	tapUtils.assert(filename.equals(central.filename), beep);
	tapUtils.assert(extra_field.equals(central.extra_field), beep);
  }

  public int getDataOffset() {
	   return offset+size;
  }

  public String toString() {
	int show=10;
	int off = getDataOffset();
	StringBuffer sb = new StringBuffer(show*3-1);
	sb.append(Integer.toString((int)data[off]&0xff,16));
	for (int i=1; i<show; i++) {
		 sb.append(" ");
		 sb.append(Integer.toString((int)data[off+i]&0xff,16));
	}

	return sb.toString();
  }
}


/**
 * Read .zip version 1.0 classes, fetch corresponding bytes by name.
 * Written on a need-to-implement basis for Pinhead, a .zip-enabled ClassLoader.
 *
 * @version		$Version$ $Date: 1996/02/22 02:56:30 $
 *
 * @author		Tom Phelps
 *
 */

public class Zippy {
  final boolean debug = false;

  public static String suffix = ".zip";

  // make the following public or provide accessors?
  public byte data[];
  public ZipDirEntry direntry[];	// assume that usually going to read all or most of files
  public ZipDirHook dirhook;

  final byte extd_local_sig[] = { 'P','K',0x07,0x08 };	// add this later as own class patterned after the above (used for subdirectories?)

  Zippy(InputStream in) {
    // suck up whole file, assuming that want most if not all classes
	ByteArrayOutputStream databuf = new ByteArrayOutputStream(50*1024);
	int chunksize = 20*1024;	// read in hunks of 20K or less
	byte buf[] = new byte[chunksize];
	int len;
	try {
		 while ((len = in.read(buf, 0, chunksize))!=-1) {
			  databuf.write(buf, 0, len);
		 }
	} catch (Exception e) {
		 System.out.println("trouble reading in zip file "+e);
		 return;
	}
	data = databuf.toByteArray();


	// DIRECTORY HOOK
	// find it (it's not fixed?)
	int offset = data.length-22;
	while (offset>=0) {
		 if (tapUtils.bytesidentical(ZipDirHook.signature,0, data,offset, ZipDirHook.signature.length)) {
			  break;
		 } else offset--;
	}
	tapUtils.assert(offset>0, "no `PK' signature");
	if (debug) System.out.println("found end-of-central-dir at "+Integer.toString(offset));

	// create it
	dirhook = new ZipDirHook(data,offset);
	if (debug) { dirhook.dump(); System.out.println(); }

	// tapUtils.assert(data.length == ...., "file length mistmatch");


	// COLLECT CENTRAL DIRECTORY ENTRIES
	direntry = new ZipDirEntry[dirhook.total_entries_central_dir];
	ZipDirEntry z;
	offset = dirhook.offset_start_central_directory;
	for (int i=0; i<direntry.length; i++) {
		 direntry[i] = z = new ZipDirEntry(data,offset);
		 tapUtils.assert(z.compression_method==0, "can't handle compressed files: "+z.filename);
		 if (debug) System.out.println(z);
		 offset += z.size();
	}
  }


  /*
   * FUNCTIONALITY
   */

  /**
   * Find the directory entry in the .zip file given the name.
   * (Possible future work: add pattern matching.)
   */
  public ZipDirEntry findName(String name) {
	ZipDirEntry z;
	for (int i=0; i<direntry.length; i++) {
	  z = direntry[i];
	  // could use regular expressions or glob patterns here
	  if (name.equals(z.filename)) return z;
	}
	return null;	// exception? no
  }

  /*
   * DIAGNOSTICS
   */

  private void datadump(int offset, int length) {
	// dump first few and last few bytes
	tapUtils.bytesdump(data,offset,length);
  }


  /**
   * Given a zip filename on the command line, dump filenames of its contents.
   * (Used for debugging.)
   */
  public static void main(String args[]) {
	Zippy zip;
	String testfile = (args.length==0?"imagebrowser.zip":args[0]);
	try {
	  zip = new Zippy(new FileInputStream(testfile));
	} catch (FileNotFoundException e) {
	  System.out.println("can't find "+testfile);
	  return;
	}

	// iterate through files dumping filename, uncompressed size, first few bytes of data
	ZipDirEntry z;
	for (int i=0; i<zip.direntry.length; i++) {
	  z = zip.direntry[i];
	  System.out.println(z);
	}

	System.out.println("... and by name");
	z = zip.findName(zip.direntry[zip.direntry.length/2].filename);
	System.out.println(z);
  }
}
