/*   -*- c++ -*-
$Header: /home/auspex/h/bair/phelps/www/java/Util/RCS/tapUtils.java,v 1.5 1996/02/17 03:27:17 phelps Exp phelps $

	generally useful functionality that should be part of Java's base classes, but isn't

	Tom Phelps

	 6 Nov	started with assert and ASCIIArmour
    22 Dec  strip HTML markup (what you really want is a widget that displays HTML!)

	1996
	14 Feb  J String <=> String (converts backslashed characters)
            (worked first time, once fixed syntax bugs!)
*/

/* NOTES *
*/

//import -- NONE
//package EDU.Berkeley.CS.Phelps.Util;

import java.net.URL;

public class tapUtils {
  static final boolean debug = true;	// set to false and recompile to zap assertion checking (I hope)

  public static void assert(boolean assertion, String errorMsg) {
	if (debug && !assertion) {
	  System.out.println("assertion violated: "+errorMsg);
	}
  }


  // remove tags, convert &xx; to character (do foreign characters later)
  public static String stripHTML(String html) {
	StringBuffer sb = new StringBuffer();
	char ch;
	boolean intag=false;
	for (int i=0; i<html.length(); i++) {
	  ch = html.charAt(i);
	  switch (ch) {
	  case '<':
		assert(!intag, "already in tag");
		intag=true;
		break;
	  case '>':
		//assert(intag, "must be in tag before closing");
		intag=false;
		break;
	  case '&':
		String tag="";
		for (i++; (ch=html.charAt(i))!=';'; i++) tag+=ch;
		if (tag.equals("gt")) sb.append('>');
		else if (tag.equals("lt")) sb.append('<');
		else if (tag.equals("amp")) sb.append('&');
		break;
	  default:
		if (!intag) sb.append(ch);
	  }
	}
	return sb.toString();
  }


  public static URL ASCIIArmour(URL rawurl) {
	URL protectedurl=null;
	try {
	  protectedurl = new URL(ASCIIArmour(rawurl.toString()));
	} catch (Exception e) {}
	return protectedurl;
  }

  public static String ASCIIArmour(String rawurl) {
    StringBuffer sb = new StringBuffer(rawurl.length());
	String meta = "%? &=/:";
	char ch;
	for (int i=0; i<rawurl.length(); i++) {
	  ch = rawurl.charAt(i);
	  if (meta.indexOf(ch)!=-1) {
		sb.append('%'); sb.append(Integer.toString(ch,16));		// convert to hex
	  } else sb.append(ch);
	}
	return sb.toString();
  }

  // no need to de-armour, I guess



  /*
   * CONVERT JSTRING <=> STRING
   */

  private static int sub2num(String s, int start, int length, int radix) {
	assert(start+length-1 < s.length(), "incomplete number at end of string "+s);
	String sub = s.substring(start,start+length);
	int num=0;
	try {
	  num = Integer.parseInt(sub,radix);
	} catch (NumberFormatException ex) {
	  assert(false, "invalid base "+radix+" number "+sub);
	}
	return num;
  }

  private static final String jmetachar = "\n\t\b\r\f\\\'\"";
  private static final String jtranschar = "ntbrf\\\'\"";

  public static String JString2String(String j) {
	StringBuffer s = new StringBuffer(j.length());	// may be shorter, but good guess

	char ch;
	int spanstart=-1;	// copy by substrings for possible efficiency gain
	for (int i=0; i<j.length(); i++) {
	  ch = j.charAt(i);
	  if (ch!='\\' || i==j.length()) {
		if (spanstart==-1) spanstart=i;
	  } else {
		// copy over span
		if (spanstart!=-1) s.append(j.substring(spanstart,i));
		spanstart=-1;

		i++; ch = j.charAt(i);
		switch (ch) {
		case 't': case 'n':	case 'b': case 'r':	case 'f':
		case '\\': case '\'': case '\"':	//"
		  ch = jmetachar.charAt(jtranschar.indexOf(ch));
		  break;
		case 'x':
		  ch = (char)sub2num(j,i+1,2,16);
		  i+=3-1;
		  break;
		case 'u':
		  ch = (char)sub2num(j,i+1,4,16);
		  i+=5-1;
		  break;
		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		  ch = (char)sub2num(j,i,3,8);
		  i+=3-1;
		  break;
		default:
		  // do nothing -- non-special characters map to themselves
		}

		s.append(ch);
	  }
	}
	// clean up any remaining span
	if (spanstart!=-1) s.append(j.substring(spanstart,j.length()));

	return s.toString();
  }


  public static String String2JString(String s) {
	StringBuffer j = new StringBuffer(s.length());	// may be longer, but good guess

	char ch;
	int meta;
	String num;
	int spanstart=-1;	// copy by substrings for possible efficiency gain
	for (int i=0; i<s.length(); i++) {
	  ch = s.charAt(i);
	  
	  if (ch>=' ' && ch<='~' && ch!='\\' && ch!='\'' && ch!='\"') {	//"
		if (spanstart==-1) spanstart=i;
	  } else {
		// escaped characters
		if (spanstart!=-1) j.append(s.substring(spanstart,i));
		spanstart=-1;

		j.append('\\');
		meta = jmetachar.indexOf(ch);
		if (meta!=-1) {
		  j.append(jtranschar.charAt(meta));
		} else if (ch<=0xFF) {
		  j.append('x');
		  num = Integer.toString(ch,16); if (ch<=0xF) j.append('0');
		  j.append(num);
		} else {
		  j.append('u');
		  num = Integer.toString(ch,16); if (ch<=0xFFF) j.append('0');
		  j.append(num);
		}
	  }
	}

	// clean up any remaining span
	if (spanstart!=-1) j.append(s.substring(spanstart,s.length()));

	return j.toString();
  }


  /*
   * byte array operations
   */

  public static boolean bytesidentical(byte b1[], int b1offset, byte b2[], int b2offset, int length) {
//System.out.println("b1offset="+b1offset+", b1.length="+b1.length+", b2offset="+b2offset+", b2.length="+b2.length+", comparison length="+length);
	assert(b1offset+length<=b1.length && b2offset+length<=b2.length, "arrays not long enough");
	for (int i=0; i<length; i++) {
		 if (b1[b1offset+i]!=b2[b2offset+i]) return false;
	}
	return true;
  }

  public static String bytetoString(byte b) {
	String m = Integer.toString(b&0xff,16);
	if (m.length()==1) m="0"+m;
	return m;
  }

  public static String bytestoString(byte b[], int offset, int length) {
	if (offset+length>b.length) length=b.length-offset;
	StringBuffer sb = new StringBuffer(length*3);

	sb.append(bytetoString(b[offset]));
	for (int i=1; i<length; i++) {
	  sb.append(" "); sb.append(bytetoString(b[offset+i]));
	}
	return sb.toString();
  }

  public static void bytesdump(byte b[], int offset, int length) {
	assert(offset<b.length, "offset >= b.length");
	if (offset<0) offset = b.length+offset;
	if (length==0) length = b.length-offset;		// 0 length => show till end
	if (offset+length>b.length) offset = b.length - length;

	// short first line
	int i=offset;
	if (i%0x10!=0) {
	  int stop = ((i/0x10)+1)*0x10;
	  System.out.print(Integer.toString(offset,8)+": ");
	  for (int j=0,spcnt=(i-((i/0x10)*0x10))*3-1; j<spcnt; j++) System.out.println(" ");
	  System.out.println(bytestoString(b,i,stop-i));
	  i=stop;
	}
	// normal lines
	// short last line
/*
	for (int i=offset; i<offset+length; i++) {
		 System.out.print(mem);
		 if (i%2==0) System.out.print(" ");
		 if (i%0x10==0) {
			  System.out.println();
			  System.out.print(Integer.toString(i,8)+": ");
		 }
	}
*/
	System.out.println();
  }



  // every class has a main() that tests its member functions
  public static void main(String args[]) {
	String url = "http://elib.cs.berkeley.edu/TR/ELIB:34/ShowPage?page=12&format=raw-gif&search=bay delta estuary|flood storm";
	System.out.println("original URL: "+url);
	System.out.println("ASCIIArmour-ized URL: "+ASCIIArmour(url));

	String s = "now\n\tis the winter \"help me\" \n";	//"
//\x2f\n";
	String j = String2JString(s);
	String res = JString2String(j);
	System.out.println("Original String => "+s+"\nEscaped => "+j+"\nreconverted => "+res);
	assert(s.equals(res), "original != reconverted");
  }
}
